/*
 * TCP Server Library Copyright (c) 2010, James Bailie.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *     * The name of James Bailie may not be used to endorse or promote
 * products derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/

/*
 * Multi-process TCP/UNIX server as a library.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/event.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <sys/un.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>

#include <pwd.h>
#include <grp.h>
#include <signal.h>

#include <syslog.h>
#include <stdarg.h>
#include <string.h>

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#define VERSION "1.11"

#define NOT_BUSY 0
#define BUSY 1
#define EMPTY -1

extern char *optarg;

int serv_fd = -1, serv_testing = 0, serv_backlog = 1000, serv_min_slaves = 3,
   serv_logging = 0, serv_slave_no = -1, serv_interval = 0,
   serv_counter = 0;

volatile int serv_killed = 0, serv_max_slaves = 25;

volatile pid_t *serv_slaves = NULL;

char *serv_interface = "", *serv_port = "1966", *serv_grp = "nobody",
   *serv_user = "nobody", *serv_busymap = NULL, *serv_root_dir = NULL,
   *serv_app_name = "server", *serv_pidfile = "server.pid",
   *serv_config_file = NULL, *serv_listen_unix = NULL;

void ( *serv_periodic_func )() = NULL;

struct passwd *serv_passwd;
struct group *serv_group;

int serv_stack_inc = 128;

void serv_set_name( char * );
void serv_init_func();
void serv_child_init_func();
void serv_service_connection();

struct serv_string
{
   int free, used;
   char *top;
   char *str;
};

void *serv_memory( int size )
{
   void *ptr;

   if ( size == 0 )
      return NULL;

   if (( ptr = malloc( size )) == NULL )
   {
      if ( serv_logging )
         syslog( LOG_WARNING, "malloc(): %m" );
      else
         fprintf( stderr, "dombey: malloc(): %s\n", strerror( errno ));

      exit( 1 );
   }

   return ptr;
}

/*
 * If we are a slave process, we set the killed flag, and when we are
 * finished servicing any current connection, we will exit.  It would be
 * nice to unlink the master process's pidfile here, but we no longer are
 * root and so cannot access /var/run/
 */

void serv_signal_handler( int signo )
{
   int i;

   if ( serv_slave_no >= 0 )
      serv_killed = 1;
   else
   {
      for( i = 0; i < serv_max_slaves; ++i )
         if ( serv_slaves[ i ] )
            kill( serv_slaves[ i ], SIGTERM );

      exit( 0 );
   }
}

/*
 * x - Do not become daemon.  Run single process iterative server.

 * l - listen on UNIX domain socket
 * m - max slaves
 * n - min free slaves
 * p - port

 * r - root directory
 * i - interface
 * u - user
 * g - group

 * f - config file
 */

void serv_set_options( int argc, char **argv )
{
   int i, specified = 0;

   while(( i = getopt( argc, argv, "xl:f:r:m:n:p:i:u:g:")) != -1 )
      switch( i )
      {
         case '?':
            exit( 1 );

         case 'f':
            serv_config_file = optarg;
            break;

         case 'r':
            serv_root_dir = optarg;
            break;

         case 'g':
            serv_grp = optarg;
            break;

         case 'i':
            ++specified;
            serv_interface = optarg;
            break;

         case 'l':
            serv_listen_unix = optarg;
            break;

         case 'm':
            serv_max_slaves = strtol( optarg, NULL, 10 );
            break;

         case 'n':
            serv_min_slaves = strtol( optarg, NULL, 10 );
            break;

         case 'p':
            ++specified;
            serv_port = optarg;
            break;

         case 'u':
            serv_user = optarg;
            break;

         case 'x':
            serv_testing = 1;
      }

   if ( specified && serv_listen_unix != NULL )
   {
      fprintf( stderr,
               "%s: the -l option cannot be present when either or both of the -i and -p options are present",
               serv_app_name );
      exit( 1 );
   }

   if ( serv_root_dir != NULL )
   {
      if ( chdir( serv_root_dir ) < 0 )
      {
         fprintf( stderr, "%s: chdir( %s ): %s\n",
                  serv_app_name, serv_root_dir, strerror( errno ));
         exit( 1 );
      }
   }

   if ( serv_min_slaves <= 0 )
   {
      fprintf( stderr, "%s: min slaves <= 0: %d\n", serv_app_name,
               serv_min_slaves );
      exit( 1 );
   }

   if ( serv_max_slaves <= 0 )
   {
      fprintf( stderr, "%s: max slaves <= 0: %d\n", serv_app_name,
               serv_max_slaves );
      exit( 1 );
   }

   if ( serv_min_slaves > serv_max_slaves )
   {
      fprintf( stderr, "%s: min slaves > max_slaves: %d > %d\n",
               serv_app_name, serv_min_slaves, serv_max_slaves );
      exit( 1 );
   }

   if (( serv_passwd = getpwnam( serv_user )) == NULL )
   {
      fprintf( stderr, "%s: user \"%s\" does not exist\n", serv_app_name,
               serv_user );
      exit( 1 );
   }

   if (( serv_group = getgrnam( serv_grp )) == NULL )
   {
      fprintf( stderr, "%s: group \"%s\" does not exist\n", serv_app_name,
               serv_grp );
      exit( 1 );
   }
}

void serv_become_daemon()
{
   int file;

   /*
    * Fork and let the parent die, continuing as child so we are not
    * a process group leader.  This is necessary for the call to setsid().
    */

   switch( fork() )
   {
      case -1:
         fprintf( stderr, "%s: fork(): %s\n", serv_app_name,
                  strerror( errno ));
         exit( 1 );

      case 0:
         break;

      default:
         exit( 0 );
   }

   fclose( stdout );
   fclose( stderr );
   fclose( stdin );

   stdin = fopen( "/dev/null", "r" );
   stdout = fopen( "/dev/null", "w" );
   stderr = fopen( "/dev/null", "w" );

   if ( stdin == NULL || stdout == NULL || stderr == NULL )
   {
      syslog( LOG_ERR, "fopen(): %m" );
      exit( 1 );
   }

   /*
    * Detach us from our controlling terminal, so job control and other
    * signals may not be sent to us from that terminal.
    */

   if ( setsid() < 0 )
   {
      syslog( LOG_ERR, "setsid(): %m" );
      exit( 1 );
   }

   /*
    * Write our pid to disk.
    */

   if ( ! serv_testing )
   {
      if (( file = open( serv_pidfile, O_WRONLY | O_CREAT | O_TRUNC,
                         S_IWUSR | S_IRUSR | S_IRGRP )) < 0 )
         syslog( LOG_WARNING, "open(): %m" );
      else
      {
         char buffer[ 16 ];

         snprintf( buffer, sizeof( buffer ), "%d", getpid() );
         write( file, buffer, strlen( buffer ));
         close( file );
      }
   }

   /*
    * We don't fork children which create files, but if we did, this will
    * make those files not world-readable by default.
    */

   umask( 2 );
}

void serv_change_identity()
{
   if ( setgid( serv_group->gr_gid ) < 0 )
      syslog( LOG_ERR, "setgid(): %m" );

   if ( setuid( serv_passwd->pw_uid ) < 0 )
      syslog( LOG_ERR, "setuid(): %m" );
}

void serv_close_on_exec( int fd )
{
   if ( fcntl( fd, F_SETFD, FD_CLOEXEC ) < 0 )
      syslog( LOG_ERR, "fcntl( F_SETFD, FD_CLOEXEC ): %m" );
}

void serv_start_listening_unix()
{
   struct sockaddr_un sa;

   if (( serv_fd = socket( PF_LOCAL, SOCK_STREAM, 0 )) < 0 )
   {
      syslog( LOG_ERR, "socket(): %m" );
      exit( 1 );
   }

   unlink( serv_listen_unix );
   bzero( &sa, sizeof( struct sockaddr_un ));
   strncpy( sa.sun_path, serv_listen_unix, sizeof( sa.sun_path ) - 1 );  /* ensures NUL-terminated. */

   if ( bind( serv_fd, ( struct sockaddr *)&sa, SUN_LEN( &sa )))
   {
      syslog( LOG_ERR, "bind( %s ): %m", serv_listen_unix );
      close( serv_fd );
      exit( 1 );
   }

   if ( chown( serv_listen_unix, serv_passwd->pw_uid, serv_group->gr_gid ) < 0 )
   {
      syslog( LOG_ERR, "chown( %s ): %m", serv_listen_unix );
      close( serv_fd );
      exit( 1 );
   }

   if ( chmod( serv_listen_unix, S_IRWXU | S_IRWXG ) < 0 )
   {
      syslog( LOG_ERR, "chmod( %s, S_IRWXU | S_IRWXG ): %m", serv_listen_unix );
      close( serv_fd );
      exit( 1 );
   }

   if ( listen( serv_fd, serv_backlog ) < 0 )
   {
      syslog( LOG_ERR, "listen(): %m" );
      close( serv_fd );
      exit( 1 );
   }

   serv_close_on_exec( serv_fd );
}

void serv_start_listening()
{
   struct addrinfo hints, *res;
   int result;

   if ( serv_listen_unix != NULL )
   {
      serv_start_listening_unix();
      return;
   }

   /*
    * This is the new, protocol-independent way of setting up a listening
    * socket.
    */

   bzero( &hints, sizeof( struct addrinfo ));
   hints.ai_flags = AI_PASSIVE;
   hints.ai_socktype = SOCK_STREAM;

   /*
    * If the user has not specified an interface, we listen on the
    * IPv6 wildcard address.
    */

   if (( result = getaddrinfo( ( *serv_interface ? serv_interface : NULL ),
                               serv_port, &hints, &res )))
   {
      syslog( LOG_ERR, "getaddrinfo(): %s", gai_strerror( result ));
      exit( 1 );
   }

   if ( res == NULL )
   {
      syslog( LOG_ERR, "getaddrinfo(): no interface found" );
      exit( 1 );
   }

   serv_fd = socket( res->ai_family, res->ai_socktype, res->ai_protocol );

   if ( serv_fd == -1 )
   {
      syslog( LOG_ERR, "socket(): %m" );
      exit( 1 );
   }

   result = 1;

   /*
    * Allow duplicate bindings, so we can have more than one
    * server process, if necessary.
    */

   if ( setsockopt( serv_fd, SOL_SOCKET, SO_REUSEADDR, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_REUSEADDR ): %m" );

   /*
    * Try and detect connections which go idle for long periods of time.
    */

   if ( setsockopt( serv_fd, SOL_SOCKET, SO_KEEPALIVE, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_KEEPALIVE ): %m" );

   /*
    * Make sure we can accept IPv4 traffic as well as IPv6, if we
    * are bound to the IPv6 wildcard address.  We are turning off
    * the option because result = 0.
    */

   result = 0;

   if ( ! *serv_interface &&
        setsockopt( serv_fd, IPPROTO_IPV6, IPV6_BINDV6ONLY, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( IPV6_BINDV6ONLY ): %m" );

   if ( bind( serv_fd, res->ai_addr, res->ai_addrlen ) < 0 )
   {
      syslog( LOG_ERR, "bind(): %m" );
      exit( 1 );
   }

   if ( res != NULL )
      freeaddrinfo( res );

   if ( listen( serv_fd, serv_backlog ) < 0 )
   {
      syslog( LOG_ERR, "listen(): %m" );
      exit( 1 );
   }

   serv_close_on_exec( serv_fd );
}

/*
 * If we are a slave, and we receive a SIGTERM, we will exit at our next
 * convenience, either the next time accept_connection() is called, after
 * we have closed the previous connection, but before we accept another, or
 * if we are already blocked in accept(), immediately, when accept() is
 * interrupted by the signal.
 */

void serv_accept_connection()
{
   struct sockaddr_storage addr;
   socklen_t len;
   int conn;

   fclose( stdin );
   fclose( stdout );

AGAIN:
   if ( serv_killed )
      exit( 0 );

   len = sizeof( struct sockaddr_storage );
   conn = accept( serv_fd, ( struct sockaddr *)&addr, &len );

   if ( conn < 0 )
   {
      if ( errno == EINTR )
         goto AGAIN;

      syslog( LOG_ERR, "accept(): %m" );
      exit( 1 );
   }

   serv_close_on_exec( conn );

   if (( stdin = fdopen( conn, "r" )) == NULL )
   {
      syslog( LOG_ERR, "fdopen(): %m" );
      exit( 1 );
   }

   if (( conn = dup( conn )) < 0 )
   {
      syslog( LOG_ERR, "dup(): %m" );
      exit( 1 );
   }

   if (( stdout = fdopen( conn, "w" )) == NULL )
   {
      syslog( LOG_ERR, "fdopen(): %m" );
      exit( 1 );
   }
}

void serv_set_name( char *name )
{
   char *ptr1, *ptr2;

   if ( name == NULL )
      return;

   serv_app_name = name;
   ptr2 = serv_pidfile = serv_memory( 14 + strlen( name ));

   for( ptr1 = "/var/run/"; *ptr1; )
      *ptr2++ = *ptr1++;

   for( ptr1 = name; *ptr1; )
      *ptr2++ = *ptr1++;

   for( ptr1 = ".pid"; *ptr1; )
      *ptr2++ = *ptr1++;

   *ptr2 = '\0';
}

void serv_service_connections()
{
   for( ; ; )
   {
      serv_busymap[ serv_slave_no ] = NOT_BUSY;
      serv_accept_connection();
      serv_busymap[ serv_slave_no ] = BUSY;

      serv_service_connection();
   }
}

void serv_init_busymap()
{
   int i;

   if (( serv_busymap = mmap( 0, serv_max_slaves, PROT_WRITE | PROT_READ,
                              MAP_ANON | MAP_SHARED, -1, 0 )) == NULL )
   {
      syslog( LOG_ERR, "mmap(): %m" );
      exit( 1 );
   }

   serv_slaves = serv_memory( sizeof( pid_t ) * serv_max_slaves );

   for( i = 0; i < serv_max_slaves; ++i )
   {
      serv_busymap[ i ] = EMPTY;
      serv_slaves[ i ] = 0;
   }
}

void serv_fork_slave( int i )
{
   pid_t pid;

   switch(( pid = fork()))
   {
      case -1:
         syslog( LOG_ERR, "fork(): %m" );
         exit( 1 );

      case 0:
         serv_slave_no = i;
         serv_child_init_func();
         serv_service_connections();
         exit( 0 );

      default:
         serv_slaves[ i ] = pid;
   }
}

void serv_periodic( void ( *func )(), int interval )
{
   serv_periodic_func = func;

   if ( interval < 0 )
      interval = 0;

   serv_interval = interval;
   serv_counter = 0;
}

void serv_manage_slaves()
{
   pid_t pid;
   int i, idle, last_idle;

   for( ; ; )
   {
      sleep( 1 );

      if ( serv_interval && serv_periodic_func != NULL && ++serv_counter == serv_interval )
      {
         serv_periodic_func();
         serv_counter = serv_interval;
      }

      idle = 0;
      last_idle = -1;

      for( i = 0; i < serv_max_slaves; ++i )
         if ( serv_busymap[ i ] == NOT_BUSY )
         {
            last_idle = i;
            ++idle;
         }

      if ( idle > serv_min_slaves )
      {
         kill( serv_slaves[ last_idle ], SIGTERM );
         --idle;
      }

      while(( pid = waitpid( -1, NULL, WNOHANG )) > 0 )
      {
         for( i = 0; i < serv_max_slaves; ++i )
            if ( pid == serv_slaves[ i ] )
            {
               serv_slaves[ i ] = 0;
               serv_busymap[ i ] = EMPTY;
               break;
            }
      }

      while( idle < serv_min_slaves )
      {
         for( i = 0; i < serv_max_slaves; ++i )
            if ( ! serv_slaves[ i ] )
            {
               serv_fork_slave( i );
               ++idle;
               break;
            }

         if ( i == serv_max_slaves )
            break;
      }
   }
}

void serv_set_signals()
{
   struct sigaction sigact;
   int *iptr;
   int sigs[] = {
      SIGPIPE, SIGHUP, SIGQUIT, SIGUSR1, SIGUSR2, SIGALRM, SIGINT, SIGTSTP, -1
   };

   /*
    * We ignore everything but SIGTERM when ! testing.  Otherwise, we
    * catch all the signals, so we may kill our children before exiting.
    * SIGTERM is set to interrupt system calls.
    */

   sigact.sa_handler = serv_signal_handler;
   sigemptyset( &sigact.sa_mask );
   sigact.sa_flags = 0;

   if ( sigaction( SIGTERM, &sigact, NULL ) < 0 )
      fprintf( stderr, "dombey: sigaction: %s\n", strerror( errno ));

   for( iptr = sigs; *iptr > 0; ++iptr )
      signal( *iptr, ( serv_testing ? serv_signal_handler : SIG_IGN ));
}

void serv_fork_slaves()
{
   int i;

   for( i = 0; i < serv_min_slaves; ++i )
      serv_fork_slave( i );
}

void serv_init_config_file( int argc, char **argv )
{
   int i;

   while(( i = getopt( argc, argv, "xl:f:r:m:n:p:i:u:g:")) != -1 )
      switch( i )
      {
         case '?':
            continue;

         case 'f':
            serv_config_file = optarg;
            break;
      }

   optreset = 1;
   optind = 1;
}

int main( int argc, char **argv )
{
   serv_init_config_file( argc, argv );
   serv_init_func();

   serv_set_options( argc, argv );
   serv_set_signals();

   openlog( serv_app_name, LOG_PID, LOG_DAEMON );
   serv_logging = 1;

   if ( !serv_testing )
      serv_become_daemon();

   serv_start_listening();
   serv_change_identity();
   serv_init_busymap();

   if ( serv_testing )
   {
      serv_child_init_func();
      serv_service_connections();
   }
   else
   {
      serv_fork_slaves();
      serv_manage_slaves();
   }

   return 0;
}
